/* $Id: vircam_imstack.c,v 1.6 2013-10-15 17:08:04 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2006 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 17:08:04 $
 * $Revision: 1.6 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <cpl.h>
#include <math.h>

#include "vircam_utils.h"
#include "vircam_mask.h"
#include "vircam_dfs.h"
#include "vircam_mods.h"
#include "vircam_fits.h"

/* Function prototypes */

static int vircam_imstack_create(cpl_plugin *) ;
static int vircam_imstack_exec(cpl_plugin *) ;
static int vircam_imstack_destroy(cpl_plugin *) ;
static int vircam_imstack_test(cpl_parameterlist *, cpl_frameset *) ;
static int vircam_imstack_save(cpl_frameset *framelist, 
				cpl_parameterlist *parlist);
static void vircam_imstack_init(void);
static void vircam_imstack_tidy(void);

/* Static global variables */

static struct {

    /* Input */

    int         method;
    int         extenum;

} vircam_imstack_config;


static struct {
    cpl_size         *labels;
    cpl_frameset     *imagelist;
    vir_fits         **images;
    cpl_frameset     *conflist;
    vir_fits         **confs;
    cpl_frameset     *catlist;
    vir_tfits        **cats;
    int              nimages;
    int              nconfs;
    vir_fits         *outfits;
    vir_fits         *outfitsc;
} ps;

static int isfirst;
static cpl_frame *product_frame = NULL;
static cpl_frame *product_conf = NULL;


static char vircam_imstack_description[] =
"vircam_imstack -- VIRCAM test jitter recipe.\n\n"
"Dither a list of frames into an output frame.\n\n"
"The program accepts the following files in the SOF:\n\n"
"    Tag                   Description\n"
"    -----------------------------------------------------------------------\n"
"    %-21s A list of images\n"
"    %-21s A list of confidence maps\n"
"    %-21s An optional list of object catalogues\n"
"\n";

/**@{*/

/**
    \ingroup testrecipelist
    \defgroup vircam_imstack vircam_imstack
    \brief Dither a series of exposures into a single frame

    \par Name: 
        vircam_imstack
    \par Purpose: 
        Dither a series of exposures into a single frame
    \par Description: 
        A load of images are dithered into a single frame.
    \par Language:
        C
    \par Parameters:
        - \b ext (int): The image extension of the input files to be done
          on this run. If all of the extensions are to be processed, then 
          this should be set to zero.
    \par Input File Types:
        The following list gives the file types that can appear in the SOF
        file. The word in bold is the DO category value.
        - \b SCIENCE_IMAGE (required): A list of images for combining
	- \b MASTER_CONF (required): A list of confidence maps, one for
	     each input image.
    \par Output File Types:
        The following lists the output data frames that are generated by 
        this recipe. The word in bold gives the PRO CATG keyword value for 
        each.
        - \b JITTERED_IMAGE_TEST: A jittered image
    \par Notes
        The extension propertylists need to have: ESO DRS XOFFDITHER (the 
	dither offset in X), ESO DRS YOFFDITHER (the dither offset in Y).
	This routine will not calculate them.
    \par Fatal Error Conditions:
        - No input images
    \par Non-Fatal Error Conditions:
        None
    \par Author:
        Jim Lewis, CASU
    \par Code Reference:
        tests/vircam_imstack.c
*/

/* Function code */

/*---------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*---------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
    cpl_recipe  *recipe = cpl_calloc(1,sizeof(*recipe));
    cpl_plugin  *plugin = &recipe->interface;
    char alldesc[SZ_ALLDESC];
    (void)snprintf(alldesc,SZ_ALLDESC,vircam_imstack_description,
		   VIRCAM_TEST_SCIENCE_RAW,VIRCAM_CAL_CONF,VIRCAM_CAL_OBJCAT);

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    VIRCAM_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "vircam_imstack",
                    "VIRCAM jitter test recipe [test]",
                    alldesc,
                    "Jim Lewis",
                    "jrl@ast.cam.ac.uk",
                    vircam_get_license(),
                    vircam_imstack_create,
                    vircam_imstack_exec,
                    vircam_imstack_destroy);

    cpl_pluginlist_append(list,plugin);

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*---------------------------------------------------------------------------*/

static int vircam_imstack_create(cpl_plugin *plugin) {
    cpl_recipe      *recipe;
    cpl_parameter   *p;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
	return(-1);

    /* Create the parameters list in the cpl_recipe object */

    recipe->parameters = cpl_parameterlist_new();

    /* Combination method */

    p = cpl_parameter_new_enum("vircam.vircam_imstack.method",
			       CPL_TYPE_INT,"Combination method",
			       "vircam.vircam_imstack",0,2,0,1);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"method");
    cpl_parameterlist_append(recipe->parameters,p);
    

    /* Extension number of input frames to use */

    p = cpl_parameter_new_range("vircam.vircam_imstack.extenum",
			        CPL_TYPE_INT,
			        "Extension number to be done, 0 == all",
			        "vircam.vircam_imstack",
			        1,0,16);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"ext");
    cpl_parameterlist_append(recipe->parameters,p);
        
    /* Get out of here */

    return(0);
}
    
    
/*---------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_imstack_exec(cpl_plugin *plugin) {
    cpl_recipe  *recipe;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
	return(-1);

    return(vircam_imstack_test(recipe->parameters,recipe->frames));
}
				
/*---------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_imstack_destroy(cpl_plugin *plugin) {
    cpl_recipe *recipe ;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
	return(-1);

    cpl_parameterlist_delete(recipe->parameters);
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_imstack_test(cpl_parameterlist *parlist, 
				cpl_frameset *framelist) {
    const char *fctid="vircam_imstack";
    int j,jst,jfn,retval,status;
    cpl_size nlab;
    cpl_parameter *p;
    

    /* Check validity of input frameset */

    if (framelist == NULL || cpl_frameset_get_size(framelist) <= 0) {
	cpl_msg_error(fctid,"Input framelist NULL or has no input data");
	return(-1);
    }

    /* Initialise some things */

    vircam_imstack_init();

    /* Get the parameters */

    p = cpl_parameterlist_find(parlist,"vircam.vircam_imstack.extenum");
    vircam_imstack_config.extenum = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,"vircam.vircam_imstack.method");
    vircam_imstack_config.method = cpl_parameter_get_int(p);

    /* Sort out raw from calib frames */

    if (vircam_dfs_set_groups(framelist) != VIR_OK) {
	cpl_msg_error(fctid,"Cannot identify RAW and CALIB frames");
	vircam_imstack_tidy();
	return(-1);
    }

    /* Get the frames frames */

    if ((ps.labels = cpl_frameset_labelise(framelist,vircam_compare_tags,
					   &nlab)) == NULL) {
	cpl_msg_error(fctid,"Cannot labelise the input frames");
	vircam_imstack_tidy();
	return(-1);
    }
    if ((ps.imagelist = vircam_frameset_subgroup(framelist,ps.labels,nlab,
						VIRCAM_TEST_SCIENCE_RAW)) == NULL) {
	cpl_msg_error(fctid,"Cannot get images in input frameset");
	vircam_imstack_tidy();
        return(-1);
    }
    ps.nimages = cpl_frameset_get_size(ps.imagelist);
    if ((ps.conflist = vircam_frameset_subgroup(framelist,ps.labels,nlab,
						VIRCAM_CAL_CONF)) == NULL) {
	cpl_msg_error(fctid,"Cannot get confidence maps in input frameset");
	vircam_imstack_tidy();
        return(-1);
    }
    ps.nconfs = cpl_frameset_get_size(ps.conflist);
    if ((ps.catlist = vircam_frameset_subgroup(framelist,ps.labels,nlab,
					       VIRCAM_CAL_OBJCAT)) == NULL) {
        cpl_msg_info(fctid,"No object catalogues found -- no matching done");
    }

    /* Now, how many image extensions do we want to do? If the extension
       number is zero, then we loop for all possible extensions. If it
       isn't then we just do the extension specified */

    vircam_exten_range(vircam_imstack_config.extenum,
		       (const cpl_frame *)cpl_frameset_get_frame(ps.imagelist,0),
		       &jst,&jfn);
    if (jst == -1 || jfn == -1) {
	cpl_msg_error(fctid,"Unable to continue");
	vircam_imstack_tidy();
	return(-1);
    }

    /* Now loop for all the extension... */

    status = VIR_OK;
    for (j = jst; j <= jfn; j++) {
	isfirst = (j == jst);

	/* Load the images */

        ps.images = vircam_fits_load_list(ps.imagelist,CPL_TYPE_FLOAT,j);
	ps.confs = vircam_fits_load_list(ps.conflist,CPL_TYPE_INT,j);
	if (ps.catlist != NULL) 
	    ps.cats = vircam_tfits_load_list(ps.catlist,j);

	/* Call the dithering module */

	cpl_msg_info(fctid,"Doing jittering for extension %" CPL_SIZE_FORMAT,
		     (cpl_size)j);
	(void)vircam_imstack(ps.images,ps.confs,ps.cats,ps.nimages,ps.nconfs,
			     5.0,5.0,vircam_imstack_config.method,0,
			     &(ps.outfits),&(ps.outfitsc),&status);
	if (status != VIR_OK) {
	    vircam_imstack_tidy();
	    return(-1);
	}

	/* Save everything */

	cpl_msg_info(fctid,"Saving combined image extension %" CPL_SIZE_FORMAT,
		     (cpl_size)j);
	retval = vircam_imstack_save(framelist,parlist);
	if (retval != 0) {
	    vircam_imstack_tidy();
	    return(-1);
	}
	freefitslist(ps.images,ps.nimages);
	freefitslist(ps.confs,ps.nconfs);
	freefits(ps.outfits);
	freefits(ps.outfitsc);
    }
    vircam_imstack_tidy();
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    The recipe data products are saved here
  @param    framelist    the input frame list
  @param    parlist      the input recipe parameter list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_imstack_save(cpl_frameset *framelist, 
				cpl_parameterlist *parlist) {
    cpl_propertylist *plist;
    const char *recipeid = "vircam_imstack";
    const char *fctid = "vircam_imstack_save";
    const char *outfile = "comb.fits";
    const char *outconf = "combconf.fits";

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list */

    if (isfirst) {

	/* Create a new product frame object and define some tags */

	product_frame = cpl_frame_new();
	cpl_frame_set_filename(product_frame,outfile);
	cpl_frame_set_tag(product_frame,VIRCAM_PRO_JITTERED_TEST);
	cpl_frame_set_type(product_frame,CPL_FRAME_TYPE_IMAGE);
	cpl_frame_set_group(product_frame,CPL_FRAME_GROUP_PRODUCT);
	cpl_frame_set_level(product_frame,CPL_FRAME_LEVEL_FINAL);

	/* Set up header for phu */

        plist = vircam_fits_get_phu(ps.outfits);
	vircam_dfs_set_product_primary_header(plist,product_frame,framelist,
					      parlist,(char *)recipeid,
					      "?Dictionary?",NULL,0);

        /* 'Save' the PHU dithered image */			 

        if (cpl_image_save(NULL,outfile,CPL_TYPE_UCHAR,plist,
			   CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
	    cpl_msg_error(fctid,"Cannot save product PHU");
	    cpl_frame_delete(product_frame);
	    return(-1);
	}
        cpl_frameset_insert(framelist,product_frame);

	/* Create a new product frame object and define some tags */

	product_conf = cpl_frame_new();
	cpl_frame_set_filename(product_conf,outconf);
	cpl_frame_set_tag(product_conf,VIRCAM_PRO_CONF_TEST);
	cpl_frame_set_type(product_conf,CPL_FRAME_TYPE_IMAGE);
	cpl_frame_set_group(product_conf,CPL_FRAME_GROUP_PRODUCT);
	cpl_frame_set_level(product_conf,CPL_FRAME_LEVEL_FINAL);

        /* 'Save' the PHU confidence map */			 

        if (cpl_image_save(NULL,outconf,CPL_TYPE_UCHAR,plist,
			   CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
	    cpl_msg_error(fctid,"Cannot save product PHU");
	    cpl_frame_delete(product_conf);
	    return(-1);
	}
        cpl_frameset_insert(framelist,product_conf);
    }

    /* Get the extension property list */

    plist = vircam_fits_get_ehu(ps.outfits);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(plist,product_frame,framelist,
					parlist,(char *)recipeid,
					"?Dictionary?",NULL);
		
    /* Now save the dithered image extension */

    if (cpl_image_save(vircam_fits_get_image(ps.outfits),outfile,
		       CPL_TYPE_FLOAT,plist,
		       CPL_IO_EXTEND) != CPL_ERROR_NONE) {
	cpl_msg_error(fctid,"Cannot save dithered image extension");
	return(-1);
    }

    /* And the confidence map */

    if (cpl_image_save(vircam_fits_get_image(ps.outfitsc),outconf,
		       CPL_TYPE_SHORT,plist,
		       CPL_IO_EXTEND) != CPL_ERROR_NONE) {
	cpl_msg_error(fctid,"Cannot save confidence map image extension");
	return(-1);
    }

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Initialise the pointers in the memory structure
 */
/*---------------------------------------------------------------------------*/

static void vircam_imstack_init(void) {
    ps.labels = NULL;
    ps.imagelist = NULL;
    ps.conflist = NULL;
    ps.catlist = NULL;
    ps.images = NULL;
    ps.confs = NULL;
    ps.cats = NULL;
    ps.outfits = NULL;
    ps.outfitsc = NULL;
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Free any allocated workspace in the memory structure
 */
/*---------------------------------------------------------------------------*/

static void vircam_imstack_tidy(void) {
    freespace(ps.labels); 
    freeframeset(ps.imagelist);
    freeframeset(ps.conflist);
    freeframeset(ps.catlist);
    freefitslist(ps.images,ps.nimages);
    freefitslist(ps.confs,ps.nconfs);
    freetfitslist(ps.cats,ps.nimages);
    freefits(ps.outfits);
    freefits(ps.outfitsc);
}

/**@}*/

/*

$Log: not supported by cvs2svn $
Revision 1.5  2012/01/16 14:44:02  jim
Fixed test recipes for cpl6 compliance

Revision 1.4  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.3  2010/06/30 12:42:00  jim
A few fixes to stop compiler compaints

Revision 1.2  2009/09/09 09:51:13  jim
modified to use new saving routines so that headers are right

Revision 1.1  2008/11/21 10:17:43  jim
New routine


*/
