/* $Id: vircam_pawsky_minus.c,v 1.1 2013-10-15 16:30:07 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2013 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:30:07 $
 * $Revision: 1.1 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cpl.h>
#include <math.h>
#include <string.h>

#include "vircam_sky.h"
#include "vircam_mask.h"
#include "vircam_fits.h"
#include "vircam_mods.h"
#include "vircam_utils.h"
#include "vircam_stats.h"
#include "vircam_wcsutils.h"

static void masksky_zeros(float **datas, unsigned char **masks, int nfiles,
			  int npts, float **zeros);
static void combine(int nfiles, int stat, int npts, float **datas, 
		    unsigned char **masks, float **skyout, 
		    unsigned char **skyout_bpm);
static void combine_mult(int nfiles, int stat, int npts, float **datas, 
			 unsigned char **masks, float *skyout);
static void domed(float *buf, int n, float *val);

extern int vircam_pawsky_minus(vir_fits **infiles, vir_fits *conf, 
			       vir_fits *objmaskfits, int nfiles, int nconfs,
			       vir_fits **skyout, int *status) {
    cpl_wcs *wcsmask,*wcsimg;
    unsigned char **masks,*sky_bpm;
    cpl_image *im,*skyim;
    double *xin,*yin,*xout,*yout,*dx,*dy,ddx,ddy;
    int nx,ny,npts,ind,i,j,*confdata,nx_mask,ny_mask,kind,ix,iy,jx,jy,*opm;
    float **datas,*sky,*zeros,val,sig;
    cpl_propertylist *plist;
    const char *fctid = "vircam_pawsky_minus";

    /* Inherited status */

    *skyout = NULL;
    if (*status != VIR_OK) 
	return(*status);

    /* If there aren't any images, then get out of here */

    if (nfiles == 0) {
	cpl_msg_error(fctid,"Sky correction impossible. No science frames");
	return(VIR_FATAL);
    }

    /* Get space for arrays */

    datas = cpl_malloc(nfiles*sizeof(float *));
    masks = cpl_malloc(nfiles*sizeof(unsigned char *));
    im = vircam_fits_get_image(infiles[0]);
    nx = cpl_image_get_size_x(im);
    ny = cpl_image_get_size_y(im);
    npts = nx*ny;

    /* Info about object mask if it exists */

    if (objmaskfits != NULL) {
	nx_mask = cpl_image_get_size_x(vircam_fits_get_image(objmaskfits));
	ny_mask = cpl_image_get_size_y(vircam_fits_get_image(objmaskfits));
	wcsmask = cpl_wcs_new_from_propertylist(vircam_fits_get_ehu(objmaskfits));
	opm = cpl_image_get_data_int(vircam_fits_get_image(objmaskfits));

	/* We need this space so that we can work out which part of the 
	   input mask is relevant for each input image */

	xin = cpl_malloc(npts*sizeof(double));
	yin = cpl_malloc(npts*sizeof(double));
	xout = cpl_malloc(npts*sizeof(double));
	yout = cpl_malloc(npts*sizeof(double));
	dx = cpl_malloc(nfiles*sizeof(double));
	dy = cpl_malloc(nfiles*sizeof(double));

	/* Initialise the input xy arrays */

	ind = 0;
	for (j = 0; j < ny; j++) {
	    for (i = 0; i < nx; i++) {
		xin[ind] = (double)(i+1);
		yin[ind++] = (double)(j+1);
	    }
	}

	/* Define an output grid for the first image and then offsets for
	   the subsequent images */

	ddx = 1.0;
	ddy = 1.0;
	for (i = 0; i < nfiles; i++) {
	    wcsimg = cpl_wcs_new_from_propertylist(vircam_fits_get_ehu(infiles[i]));
	    if (i == 0) 
		vircam_xytoxy_list(wcsimg,wcsmask,npts,xin,yin,xout,yout);
	    vircam_xytoxy_list(wcsimg,wcsmask,1,&ddx,&ddy,dx+i,dy+i);
	    dx[i] -= xout[0];
	    dy[i] -= yout[0];
	    cpl_wcs_delete(wcsimg);
	}
	cpl_wcs_delete(wcsmask);
    }

    /* Loop for each file and get a reference for the data array. Get
       some space for a mask that will cover both the zeroed pixels in the
       confidence maps and any object mask that might be used */

    for (i = 0; i < nfiles; i++) {
	im = vircam_fits_get_image(infiles[i]);
	datas[i] = cpl_image_get_data_float(im);
	masks[i] = cpl_calloc(npts,sizeof(unsigned char));
	if (i == 0) 
	    confdata = cpl_image_get_data_int(vircam_fits_get_image(conf));

	/* Now loop for each pixel. If the confidence map is zero here then
	   mark this as a bad pixel. If not, then look at the master mask
	   and see which is its closest pixel to the current one. NB: the 
	   -0.5 in the index is a combination of +0.5 to do a nint and a -1
	   to take account of the fact that arrays start from index zero. */

	for (jy = 0; jy < ny; jy++) {
	    for (jx = 0; jx < nx; jx++) {
		ind = jy*nx + jx;
		if (confdata[ind] == 0) {
		    masks[i][ind] = 1;
		} else if (objmaskfits != NULL) {
		    kind = (int)(yout[ind] + dy[i] - 0.5)*nx_mask + 
			(int)(xout[ind] + dx[i] - 0.5);
		    masks[i][ind] = opm[kind];
		}
	    }
	}
    }

    /* Do an intermediate tidy */

    if (objmaskfits != NULL) {
	freespace(xin);
	freespace(yin);
	freespace(xout);
	freespace(yout);
	freespace(dx);
	freespace(dy);
    }

    /* Offset the DC level */
    
    masksky_zeros(datas,masks,nfiles,npts,&zeros);

    /* Right, combine these now */

    combine(nfiles,0,npts,datas,masks,&sky,&sky_bpm);
    
    /* Wrap the sky and reject the pixels that had no sky contribution */

    skyim = cpl_image_wrap_float(nx,ny,sky);
    for (i = 0; i < npts; i++) {
	iy = i/nx + 1;
	ix = npts - (iy-1)*nx;
	if (sky_bpm[i])
	    cpl_image_reject(skyim,(cpl_size)ix,(cpl_size)iy);
    }
    freespace(sky_bpm);
    *skyout = vircam_fits_wrap(skyim,infiles[0],NULL,NULL);
    plist = vircam_fits_get_ehu(*skyout);
    if (objmaskfits != NULL) {
	cpl_propertylist_update_string(plist,"ESO DRS MASKUSED",
				       vircam_fits_get_filename(objmaskfits));
	cpl_propertylist_set_comment(plist,"ESO DRS MASKUSED",
				     "Object masked used to make sky");
    }
    cpl_propertylist_update_string(plist,"ESO DRS SKYALGO","pawsky_minus");
    cpl_propertylist_set_comment(plist,"ESO DRS SKYALGO",
				 "Sky estimation algorithm");
    vircam_prov(plist,infiles,nfiles);

    /* Fill in the bits with no sky contribution */

    vircam_inpaint(*skyout,64,status);

    /* Do the multiple combine and subtraction */

    combine_mult(nfiles,0,npts,datas,masks,sky);

    /* Now add the DC level back onto the data */

    for (i = 0; i < nfiles; i++) {
	vircam_qmedsig(datas[i],masks[i],(long)npts,3.0,1,-1000.0,65535.0,&val,
		       &sig);
	val = zeros[i] - val;
	for (j = 0; j < npts; j++) 
	    datas[i][j] += val;
    }
    freespace(zeros);

    /* Add a little something to the header and start cleaning up */

    for (i = 0; i < nfiles; i++) {
	plist = vircam_fits_get_ehu(infiles[i]);
	cpl_propertylist_update_string(plist,"ESO DRS SKYSUB",
				       "Done with pawsky_minus");
	freespace(masks[i]);
    }
    freespace(masks);
    freespace(datas);
    return(VIR_OK);
}

static void masksky_zeros(float **datas, unsigned char **masks, int nfiles,
			  int npts, float **zeros) {
    int i,j;
    float sig,off,medval;

    /* Get some space for the zeros */

    *zeros = cpl_malloc(nfiles*sizeof(float));

    /* Loop for each input image and get the background median */

    for (i = 0; i < nfiles; i++) 
	vircam_qmedsig(datas[i],masks[i],(long)npts,3.0,1,-1000.0,
		       65535.0,*zeros+i,&sig);

    /* Get the median value of the array */

    medval = vircam_med(*zeros,NULL,nfiles);

    /* Now work out the offset for each image and subtract it */

    for (i = 0; i < nfiles; i++) {
	off = medval - (*zeros)[i];
	for (j = 0; j < npts; j++)
	    datas[i][j] += off;
    }
}

static void combine(int nfiles, int stat, int npts, float **datas, 
		    unsigned char **masks, float **skyout, 
		    unsigned char **skyout_bpm) {
    int i,j,n;
    float *buf;

    /* Get workspace for output arrays */

    *skyout = cpl_malloc(npts*sizeof(float));
    *skyout_bpm = cpl_malloc(npts*sizeof(unsigned char));
    buf = cpl_malloc(nfiles*sizeof(float));

    /* Loop now for each input pixel and form the median */

    for (j = 0; j < npts; j++) {
	n = 0;
	for (i = 0; i < nfiles; i++) {
	    if (masks[i][j] == 0) 
		buf[n++] = datas[i][j];
	}
	if (n == 0) {
	    (*skyout)[j] = 0.0;
	    (*skyout_bpm)[j] = 1;
	} else {
	    if (stat == 0) 
		domed(buf,n,(*skyout)+j);
	    else 
		(*skyout)[j] = vircam_mean(buf,NULL,n);
	    (*skyout_bpm)[j] = 0;
	}
    }

    /* Free some workspace and get out of here */

    freespace(buf);
    return;
}

static void combine_mult(int nfiles, int stat, int npts, float **datas, 
			 unsigned char **masks, float *skyout) {
    int i,j,k,n;
    float *buf,*val;

    /* Get some memory */

    buf = cpl_malloc(nfiles*sizeof(float));
    val = cpl_malloc(nfiles*sizeof(float));

    /* Loop now for each input pixel */

    for (j = 0; j < npts; j++) {

	/* Loop for each file. If the current pixel in the current file
	   was flagged, then the sky for that pixel will be the same as
	   the general sky we did before. If not, then remove that file's
	   pixel and recalculate the sky */

	for (i = 0; i < nfiles; i++) {
	    if (masks[i][j] != 0) {
		val[i] = skyout[j];
	    } else {
		n = 0;
		for (k = 0; k < nfiles; k++) {
		    if (k == i || masks[k][j] != 0)
			continue;
		    buf[n++] = datas[k][j];
		}
		if (n == 0) {
		    val[i] = skyout[j];
		} else {
		    if (stat == 0) 
			domed(buf,n,val+i);
		    else 
			val[i] = vircam_mean(buf,NULL,n);
		}
	    }
	}

	/* Now correct these pixels in the input images */

	for (i = 0; i < nfiles; i++) 
	    datas[i][j] -= val[i];
    }

    /* Free some workspace and get out of here */

    freespace(buf);
    freespace(val);
    return;
}


static void domed(float *buf, int n, float *val) {
    int is_even,n2;

    is_even = ! (n & 1);
    vircam_sort(&buf,n,1);
    n2 = n/2;
    if (is_even) {
	*val = 0.5*(buf[n2-1] + buf[n2]);
    } else {
	if (n <= 5) 
	    *val = buf[n2];
	else
	    *val = 0.5*buf[n2] + 0.25*(buf[n2-1] + buf[n2+1]);
    }
}

/*

$Log: not supported by cvs2svn $

*/

